import lightning

from pvnet_summation.data.datamodule import PresavedDataModule
from pvnet_summation.optimizers import AdamW
from pvnet_summation.training.lightning_module import PVNetSummationLightningModule


def test_model_trainer_fit(model, presaved_samples_dir):

    datamodule = PresavedDataModule(
        sample_dir=presaved_samples_dir,
        batch_size=2,
        num_workers=0,
        prefetch_factor=None,
    )

    model_module = PVNetSummationLightningModule(
        model=model,
        optimizer=AdamW()
    )

    trainer = lightning.pytorch.trainer.trainer.Trainer(fast_dev_run=True)
    trainer.fit(model=model_module, datamodule=datamodule)



"""Neural network architecture based on dense layers applied independently at each horizon"""


import torch
import torch.nn.functional as F
from torch import nn

from pvnet_summation.data.datamodule import SumTensorBatch
from pvnet_summation.models.base_model import BaseModel


class HorizonDenseModel(BaseModel):
    """Neural network architecture based on dense layers applied independently at each horizon.
    """

    def __init__(
        self,
        output_quantiles: list[float] | None,
        num_input_locations: int,
        input_quantiles: list[float] | None,
        history_minutes: int,
        forecast_minutes: int,
        interval_minutes: int,
        output_network: torch.nn.Module,        
        predict_difference_from_sum: bool = False,
        use_horizon_encoding: bool = False,
        use_solar_position: bool = False,
        force_non_crossing: bool = False,
        beta: float = 3,
    ):
        """Neural network architecture based on dense layers applied independently at each horizon.

        Args:
            output_quantiles: A list of float (0.0, 1.0) quantiles to predict values for. If set to
                None the output is a single value.
            num_input_locations: The number of input locations (e.g. number of GSPs)
            input_quantiles: A list of float (0.0, 1.0) quantiles which PVNet predicts for. If set 
                to None we assume PVNet predicts a single value
            history_minutes (int): Length of the GSP history period in minutes
            forecast_minutes (int): Length of the GSP forecast period in minutes
            interval_minutes: The interval in minutes between each timestep in the data
            output_network: A partially instantiated pytorch Module class used top predict the 
                outturn at each horizon.
            predict_difference_from_sum: Whether to predict the difference from the sum of locations
                else the total is predicted directly
            use_horizon_encoding: Whether to use the forecast horizon as an input feature
            use_solar_position: Whether to use the solar coordinates as input features
            force_non_crossing: If predicting quantile, whether to predict the quantiles other than
                the median by predicting the distance between them and integrating.
            beta: If using force_non_crossing, the beta value to use in the softplus activation
        """

        super().__init__(
            output_quantiles, 
            num_input_locations,
            input_quantiles,
            history_minutes,
            forecast_minutes,
            interval_minutes,
        )

        if force_non_crossing:
            assert self.use_quantile_regression

        self.use_horizon_encoding = use_horizon_encoding
        self.predict_difference_from_sum = predict_difference_from_sum
        self.force_non_crossing = force_non_crossing
        self.beta = beta
        self.use_solar_position = use_solar_position

        in_features = 1 if self.input_quantiles is None else len(self.input_quantiles)
        in_features = in_features * self.num_input_locations

        if use_horizon_encoding:
            in_features += 1

        if use_solar_position:
            in_features += 2

        out_features = (len(self.output_quantiles) if self.use_quantile_regression else 1)

        model = output_network(in_features=in_features, out_features=out_features)

        # Add linear layer if predicting difference from sum
        # - This allows difference to be positive or negative
        # Also add linear layer if we are applying force_non_crossing since a softplus will be used
        if predict_difference_from_sum or force_non_crossing:
            model = nn.Sequential(
                model,
                nn.Linear(out_features, out_features),
            )
        
        self.model = model


    def forward(self, x: SumTensorBatch) -> torch.Tensor:
        """Run model forward"""

        # x["pvnet_outputs"] has shape [batch, locs, horizon, (quantile)]
        batch_size = x["pvnet_outputs"].shape[0]
        x_in = torch.swapaxes(x["pvnet_outputs"], 1, 2) # -> [batch, horizon, locs, (quantile)]
        x_in = torch.flatten(x_in, start_dim=2) # -> [batch, horizon, locs*(quantile)]

        if self.use_horizon_encoding:
            horizon_encoding = torch.linspace(
                start=0,
                end=1,
                steps=self.forecast_len,
                device=x_in.device,
                dtype=x_in.dtype,
            )
            horizon_encoding = horizon_encoding.tile((batch_size,1)).unsqueeze(-1)
            x_in = torch.cat([x_in, horizon_encoding], dim=2)

        if self.use_solar_position:
            x_in = torch.cat(
                [x_in, x["azimuth"].unsqueeze(-1), x["elevation"].unsqueeze(-1)],
                dim=2
            )

        x_in = torch.flatten(x_in, start_dim=0, end_dim=1) # -> [batch*horizon, features]

        out = self.model(x_in)
        out = out.view(batch_size, *self.output_shape) # -> [batch, horizon, (quantile)]

        if self.force_non_crossing:
            
            # Get the prediction of the median
            idx = self.output_quantiles.index(0.5)
            if self.predict_difference_from_sum:
                loc_sum = self.sum_of_locations(x).unsqueeze(-1)
                y_median = loc_sum + out[..., idx:idx+1]
            else:
                y_median = out[..., idx:idx+1]

            # These are the differences between the remaining quantiles
            dy_below = F.softplus(out[..., :idx], beta=self.beta)
            dy_above = F.softplus(out[..., idx+1:], beta=self.beta)

            # Find the absolute value of the quantile predictions from the differences
            y_below = []
            y = y_median
            for i in range(dy_below.shape[-1]):
                # We detach y to avoid the gradients caused by errors from one quantile 
                # prediction  flowing back to affect the other quantile predictions.
                # For example if the 0.9 quantile prediction was too low, we don't want the
                # gradient to pull the 0.5 quantile prediction higher to compensate.
                y = y.detach() - dy_below[..., i:i+1]
                y_below.append(y)

            y_above = []
            y = y_median
            for i in range(dy_above.shape[-1]):
                y = y.detach() + dy_above[..., i:i+1]
                y_above.append(y)

            # Compile the quantile predictions in the correct order
            out = torch.cat(y_below[::-1] + [y_median,] + y_above, dim=-1)

        else:

            if self.predict_difference_from_sum:
                loc_sum = self.sum_of_locations(x)

                if self.use_quantile_regression:
                    loc_sum = loc_sum.unsqueeze(-1)

                out = loc_sum + out

        # Use leaky relu as a soft clip to 0
        return F.leaky_relu(out, negative_slope=0.01)


"""Simple model which only uses outputs of PVNet for all GSPs"""

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

from pvnet_summation.data.datamodule import SumTensorBatch
from pvnet_summation.models.base_model import BaseModel


class DenseModel(BaseModel):
    """Neural network architecture based on naive dense layers

    This model flattens all the features into a 1D vector before feeding them into the sub network
    """

    def __init__(
        self,
        output_quantiles: list[float] | None,
        num_input_locations: int,
        input_quantiles: list[float] | None,
        history_minutes: int,
        forecast_minutes: int,
        interval_minutes: int,
        output_network: torch.nn.Module,        
        predict_difference_from_sum: bool = False,
    ):
        """Neural network architecture based on naive dense layers

        """

        super().__init__(
            output_quantiles, 
            num_input_locations,
            input_quantiles,
            history_minutes,
            forecast_minutes,
            interval_minutes,
        )

        self.predict_difference_from_sum = predict_difference_from_sum

        self.model = output_network(
            in_features=np.prod(self.input_shape),
            out_features=self.num_output_features,
        )

        # Add linear layer if predicting difference from sum
        # This allows difference to be positive or negative
        if predict_difference_from_sum:
            self.model = nn.Sequential(
                self.model, 
                nn.Linear(self.num_output_features, self.num_output_features),
            )

    def forward(self, x: SumTensorBatch) -> torch.Tensor:
        """Run model forward"""

        x_in = torch.flatten(x["pvnet_outputs"], start_dim=1)
        out = self.model(x_in)

        if self.use_quantile_regression:
            # Shape: [batch_size, seq_length * num_quantiles]
            out = out.reshape(out.shape[0], self.forecast_len, len(self.output_quantiles))

        if self.predict_difference_from_sum:
            loc_sum = self.sum_of_locations(x)

            if self.use_quantile_regression:
                loc_sum = loc_sum.unsqueeze(-1)

            out = F.leaky_relu(loc_sum + out)

        return out



"""Training"""
import logging
import os

import hydra
import torch
from lightning.pytorch import Callback, Trainer, seed_everything
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import Logger, WandbLogger
from ocf_data_sampler.torch_datasets.utils.torch_batch_utils import (
    batch_to_tensor,
    copy_batch_to_device,
)
from omegaconf import DictConfig, OmegaConf
from pvnet.models import BaseModel as PVNetBaseModel
from tqdm import tqdm

from pvnet_summation.data.datamodule import PresavedDataModule, StreamedDataModule
from pvnet_summation.utils import (
    DATAMODULE_CONFIG_NAME,
    FULL_CONFIG_NAME,
    MODEL_CONFIG_NAME,
    create_pvnet_model_config,
)

log = logging.getLogger(__name__)


def resolve_monitor_loss(output_quantiles: list | None) -> str:
    """Return the desired metric to monitor based on whether quantile regression is being used.

    Adds the option to use
        monitor: "${resolve_monitor_loss:${model.model.output_quantiles}}"
    in early stopping and model checkpoint callbacks so the callbacks config does not need to be
    modified depending on whether quantile regression is being used or not.
    """
    if output_quantiles is None:
        return "MAE/val"
    else:
        return "quantile_loss/val"


OmegaConf.register_new_resolver("resolve_monitor_loss", resolve_monitor_loss)


def train(config: DictConfig) -> None:
    """Contains training pipeline.

    Instantiates all PyTorch Lightning objects from config.

    Args:
        config (DictConfig): Configuration composed by Hydra.
    """

    # Get the PVNet model
    pvnet_model = PVNetBaseModel.from_pretrained(
        model_id=config.datamodule.pvnet_model.model_id,
        revision=config.datamodule.pvnet_model.revision
    )
    
    # Enable adding new keys to config
    OmegaConf.set_struct(config, False)
    # Set summation model parameters to align with the input PVNet model
    config.model.model.history_minutes = pvnet_model.history_minutes
    config.model.model.forecast_minutes = pvnet_model.forecast_minutes
    config.model.model.interval_minutes = pvnet_model.interval_minutes
    config.model.model.num_input_locations = len(pvnet_model.location_id_mapping)
    config.model.model.input_quantiles = pvnet_model.output_quantiles
    OmegaConf.set_struct(config, True)

    # Set seed for random number generators in pytorch, numpy and python.random
    if "seed" in config:
        seed_everything(config.seed, workers=True)

    # Compute and save the PVNet predictions before training the summation model
    save_dir = (
        f"{config.sample_save_dir}/{config.datamodule.pvnet_model.model_id}"
        f"/{config.datamodule.pvnet_model.revision}"
    )

    if os.path.isdir(save_dir):
        log.info(
            f"PVNet output directory already exists: {save_dir}\n"
            "Skipping saving new outputs. The existing saved outputs will be loaded."
        )
    else:
        log.info(f"Saving PVNet outputs to {save_dir}")

        # Move to device and disable gradients for inference
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        pvnet_model.to(device).requires_grad_(False)

        os.makedirs(f"{save_dir}/train")
        os.makedirs(f"{save_dir}/val")

        pvnet_data_config_path = f"{save_dir}/pvnet_data_config.yaml"

        data_source_paths = OmegaConf.to_container(
            config.datamodule.data_source_paths, 
            resolve=True,
        )

        create_pvnet_model_config(
            save_path=pvnet_data_config_path,
            repo=config.datamodule.pvnet_model.model_id,
            commit=config.datamodule.pvnet_model.revision,
            data_source_paths=data_source_paths,
        )

        datamodule = StreamedDataModule(
            configuration=pvnet_data_config_path,
            num_workers=config.datamodule.num_workers,
            prefetch_factor=config.datamodule.prefetch_factor,
            train_period=config.datamodule.train_period,
            val_period=config.datamodule.val_period,
            persistent_workers=False,
            seed=config.datamodule.seed,
            dataset_pickle_dir=config.datamodule.dataset_pickle_dir,
        )

        datamodule.setup()

        for dataloader_func, max_num_samples, split in [
            (datamodule.train_dataloader, config.datamodule.max_num_train_samples, "train",),
            (datamodule.val_dataloader, config.datamodule.max_num_val_samples, "val"),
        ]:
            
            log.info(f"Saving {split} outputs")
            dataloader = dataloader_func(shuffle=True)

            # If max_num_samples set to None use all samples
            max_num_samples = max_num_samples or len(dataloader)

            for i, sample in tqdm(zip(range(max_num_samples), dataloader), total=max_num_samples):
                # Run PVNet inputs though model
                x = copy_batch_to_device(batch_to_tensor(sample["pvnet_inputs"]), device)
                pvnet_outputs = pvnet_model(x).detach().cpu()

                # Create version of sample without the PVNet inputs and save
                sample_to_save = {k: v.clone() for k, v in sample.items() if k!="pvnet_inputs"}

                sample_to_save["pvnet_outputs"] = pvnet_outputs
                torch.save(sample_to_save, f"{save_dir}/{split}/{i:06}.pt")
            
            del dataloader

        datamodule.teardown()


    datamodule = PresavedDataModule(
        sample_dir=save_dir,
        batch_size=config.datamodule.batch_size,
        num_workers=config.datamodule.num_workers,
        prefetch_factor=config.datamodule.prefetch_factor,
        persistent_workers=config.datamodule.persistent_workers,
    )

    # Init lightning loggers
    loggers: list[Logger] = []
    if "logger" in config:
        for _, lg_conf in config.logger.items():
            loggers.append(hydra.utils.instantiate(lg_conf))

    # Init lightning callbacks
    callbacks: list[Callback] = []
    if "callbacks" in config:
        for _, cb_conf in config.callbacks.items():
            callbacks.append(hydra.utils.instantiate(cb_conf))

    # Align the wandb id with the checkpoint path
    # - only works if wandb logger and model checkpoint used
    # - this makes it easy to push the model to huggingface
    use_wandb_logger = False
    for logger in loggers:
        if isinstance(logger, WandbLogger):
            use_wandb_logger = True
            wandb_logger = logger
            break

    # Set the output directory based in the wandb-id of the run
    if use_wandb_logger:
        for callback in callbacks:
            if isinstance(callback, ModelCheckpoint):
                # Calling the .experiment property instantiates a wandb run
                wandb_id = wandb_logger.experiment.id

                # Save the run results to the expected parent folder but with the folder name
                # set by the wandb ID
                save_dir = f"{os.path.dirname(callback.dirpath)}/{wandb_id}"

                callback.dirpath = save_dir
                
                # Save the model config
                os.makedirs(save_dir, exist_ok=True)
                OmegaConf.save(config.model, f"{save_dir}/{MODEL_CONFIG_NAME}")

                # Save the datamodule config
                OmegaConf.save(config.datamodule, f"{save_dir}/{DATAMODULE_CONFIG_NAME}")

                # Save the full hydra config to the output directory and to wandb
                OmegaConf.save(config, f"{save_dir}/{FULL_CONFIG_NAME}")
                wandb_logger.experiment.save(f"{save_dir}/{FULL_CONFIG_NAME}", base_path=save_dir)


    # Init lightning model
    model = hydra.utils.instantiate(config.model)
                
    trainer: Trainer = hydra.utils.instantiate(
        config.trainer,
        logger=loggers,
        _convert_="partial",
        callbacks=callbacks,
    )

    # Train the model completely
    trainer.fit(model=model, datamodule=datamodule)


"""Utils"""
import logging

import rich.syntax
import rich.tree
import yaml
from lightning.pytorch.utilities import rank_zero_only
from omegaconf import DictConfig, OmegaConf
from pvnet.models.base_model import BaseModel as PVNetBaseModel

logger = logging.getLogger(__name__)


PYTORCH_WEIGHTS_NAME = "model_weights.safetensors"
MODEL_CONFIG_NAME = "model_config.yaml"
DATAMODULE_CONFIG_NAME = "datamodule_config.yaml"
FULL_CONFIG_NAME =  "full_experiment_config.yaml"
MODEL_CARD_NAME = "README.md"



def maybe_apply_debug_mode(config: DictConfig) -> None:
    """Check if debugging run is requested and force debug-frendly configuration

    Controlled by main config file

    Modifies DictConfig in place.

    Args:
        config (DictConfig): Configuration composed by Hydra.
    """

    # Enable adding new keys to config
    OmegaConf.set_struct(config, False)

    # Force debugger friendly configuration if <config.trainer.fast_dev_run=True>
    if config.trainer.get("fast_dev_run"):
        logger.info("Forcing debugger friendly configuration! <config.trainer.fast_dev_run=True>")
        # Debuggers don't like GPUs or multiprocessing
        if config.trainer.get("gpus"):
            config.trainer.gpus = 0
        if config.datamodule.get("pin_memory"):
            config.datamodule.pin_memory = False
        if config.datamodule.get("num_workers"):
            config.datamodule.num_workers = 0
        if config.datamodule.get("prefetch_factor"):
            config.datamodule.prefetch_factor = None

    # Disable adding new keys to config
    OmegaConf.set_struct(config, True)


@rank_zero_only
def print_config(
    config: DictConfig,
    fields: tuple[str, ...] = (
        "trainer",
        "model",
        "datamodule",
        "callbacks",
        "logger",
        "seed",
    ),
    resolve: bool = True,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.

    Args:
        config (DictConfig): Configuration composed by Hydra.
        fields (tuple[str, ...], optional): Determines which main fields from config will
        be printed and in what order.
        resolve (bool, optional): Whether to resolve reference fields of DictConfig.
    """

    style = "dim"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    for field in fields:
        branch = tree.add(field, style=style, guide_style=style)

        config_section = config.get(field)
        branch_content = str(config_section)
        if isinstance(config_section, DictConfig):
            branch_content = OmegaConf.to_yaml(config_section, resolve=resolve)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    rich.print(tree)

def populate_config_with_data_data_filepaths(config: dict, data_source_paths: dict) -> dict:
    """Populate the data source filepaths in the config

    Args:
        config: The data config
        data_source_paths: A dictionary of data paths for the different input sources
    """

    # Replace the GSP data path
    config["input_data"]["gsp"]["zarr_path"] =  data_source_paths["gsp"]

    # Replace satellite data path if using it
    if "satellite" in config["input_data"]:
        if config["input_data"]["satellite"]["zarr_path"] != "":
            config["input_data"]["satellite"]["zarr_path"] = data_source_paths["satellite"]

    # NWP is nested so much be treated separately
    if "nwp" in config["input_data"]:
        nwp_config = config["input_data"]["nwp"]
        for nwp_source in nwp_config.keys():
            provider = nwp_config[nwp_source]["provider"]
            assert provider in data_source_paths["nwp"], f"Missing NWP path: {provider}"
            nwp_config[nwp_source]["zarr_path"] = data_source_paths["nwp"][provider]

    return config


def create_pvnet_model_config(
    save_path: str, 
    repo: str, 
    commit: str, 
    data_source_paths: dict,
) -> None:
    """Create the data config needed to run the PVNet model"""
    data_config_path = PVNetBaseModel.get_data_config(repo, revision=commit)

    with open(data_config_path) as file:
        data_config = yaml.load(file, Loader=yaml.FullLoader)

    data_config = populate_config_with_data_data_filepaths(data_config, data_source_paths)

    with open(save_path, "w") as file:
        yaml.dump(data_config, file, default_flow_style=False)


"""Base model for all PVNet submodels"""
import logging
import os
import shutil
import time
from importlib.metadata import version
from math import prod
from pathlib import Path

import hydra
import torch
import yaml
from huggingface_hub import ModelCard, ModelCardData, snapshot_download
from huggingface_hub.hf_api import HfApi
from safetensors.torch import load_file, save_file

from pvnet_summation.data.datamodule import SumTensorBatch
from pvnet_summation.utils import (
    DATAMODULE_CONFIG_NAME,
    FULL_CONFIG_NAME,
    MODEL_CARD_NAME,
    MODEL_CONFIG_NAME,
    PYTORCH_WEIGHTS_NAME,
)


def santize_datamodule(config: dict) -> dict:
    """Create new datamodule config which only keeps the details required for inference"""
    return {"pvnet_model": config["pvnet_model"]}

def download_from_hf(
    repo_id: str,
    filename: str | list[str],
    revision: str,
    cache_dir: str | None,
    force_download: bool,
    max_retries: int = 5,
    wait_time: int = 10,
    token: bool | str | None = None,
) -> str | list[str]:
    """Tries to download one or more files from HuggingFace up to max_retries times.

    Args:
        repo_id: HuggingFace repo ID
        filename: Name of the file(s) to download
        revision: Specific model revision
        cache_dir: Cache directory
        force_download: Whether to force a new download
        max_retries: Maximum number of retry attempts
        wait_time: Wait time (in seconds) before retrying
        token:
            HF authentication token. If True, the token is read from the HuggingFace config folder.
            If a string, it is used as the authentication token.

    Returns:
        The local file path of the downloaded file(s)
    """
    for attempt in range(1, max_retries + 1):
        try:
            save_dir = snapshot_download(
                repo_id=repo_id,
                allow_patterns=filename,
                revision=revision,
                cache_dir=cache_dir,
                force_download=force_download,
                token=token,
            )

            if isinstance(filename, list):
                return [f"{save_dir}/{f}" for f in filename]
            else:
                return f"{save_dir}/{filename}"
        
        except Exception as e:
            if attempt == max_retries:
                raise Exception(
                    f"Failed to download {filename} from {repo_id} after {max_retries} attempts."
                ) from e
            logging.warning(
                (
                    f"Attempt {attempt}/{max_retries} failed to download {filename} "
                    f"from {repo_id}. Retrying in {wait_time} seconds..."
                )
            )
            time.sleep(wait_time)


class HuggingfaceMixin:
    """Mixin for saving and loading model to and from huggingface"""

    @classmethod
    def from_pretrained(
        cls,
        model_id: str,
        revision: str,
        cache_dir: str | None = None,
        force_download: bool = False,
        strict: bool = True,
        token: bool | str | None = None,
    ) -> "BaseModel":
        """Load Pytorch pretrained weights and return the loaded model."""

        if os.path.isdir(model_id):
            print("Loading model from local directory")
            model_file = f"{model_id}/{PYTORCH_WEIGHTS_NAME}"
            config_file = f"{model_id}/{MODEL_CONFIG_NAME}"
        else:
            print("Loading model from huggingface repo")

            model_file, config_file = download_from_hf(
                repo_id=model_id,
                filename=[PYTORCH_WEIGHTS_NAME, MODEL_CONFIG_NAME],
                revision=revision,
                cache_dir=cache_dir,
                force_download=force_download,
                max_retries=5,
                wait_time=10,
                token=token,
            )

        with open(config_file, "r") as f:
            model = hydra.utils.instantiate(yaml.safe_load(f))

        state_dict = load_file(model_file)
        model.load_state_dict(state_dict, strict=strict)  # type: ignore
        model.eval()  # type: ignore

        return model
    
    @classmethod
    def get_datamodule_config(
        cls,
        model_id: str,
        revision: str,
        cache_dir: str | None = None,
        force_download: bool = False,
        token: bool | str | None = None,
    ) -> str:
        """Load data config file."""
        if os.path.isdir(model_id):
            print("Loading datamodule config from local directory")
            datamodule_config_file = os.path.join(model_id, DATAMODULE_CONFIG_NAME)
        else:
            print("Loading datamodule config from huggingface repo")
            datamodule_config_file = download_from_hf(
                repo_id=model_id,
                filename=DATAMODULE_CONFIG_NAME,
                revision=revision,
                cache_dir=cache_dir,
                force_download=force_download,
                max_retries=5,
                wait_time=10,
                token=token,
            )

        return datamodule_config_file

    def _save_model_weights(self, save_directory: str) -> None:
        """Save weights from a Pytorch model to a local directory."""
        save_file(self.state_dict(), f"{save_directory}/{PYTORCH_WEIGHTS_NAME}")

    def save_pretrained(
        self,
        save_directory: str,
        model_config: dict,
        wandb_repo: str,
        wandb_id: str,
        card_template_path: str,
        datamodule_config_path,
        experiment_config_path: str | None = None,
        hf_repo_id: str | None = None,
        push_to_hub: bool = False,
    ) -> None:
        """Save weights in local directory or upload to huggingface hub.

        Args:
            save_directory:
                Path to directory in which the model weights and configuration will be saved.
            model_config (`dict`):
                Model configuration specified as a key/value dictionary.
            wandb_repo: Identifier of the repo on wandb.
            wandb_id: Identifier of the model on wandb.
            datamodule_config_path:
                The path to the datamodule config.
            card_template_path: Path to the HuggingFace model card template. Defaults to card in
                PVNet library if set to None.
            experiment_config_path:
                The path to the full experimental config.
            hf_repo_id:
                ID of your repository on the Hub. Used only if `push_to_hub=True`. Will default to
                the folder name if not provided.
            push_to_hub (`bool`, *optional*, defaults to `False`):
                Whether or not to push your model to the HuggingFace Hub after saving it.
        """

        save_directory = Path(save_directory)
        save_directory.mkdir(parents=True, exist_ok=True)

        # Save model weights/files
        self._save_model_weights(save_directory)

        # Save the model config
        if isinstance(model_config, dict):
            with open(save_directory / MODEL_CONFIG_NAME, "w") as outfile:
                yaml.dump(model_config, outfile, sort_keys=False, default_flow_style=False)

        # Sanitize and save the datamodule config
        with open(datamodule_config_path) as cfg:
            datamodule_config = yaml.load(cfg, Loader=yaml.FullLoader)

        datamodule_config = santize_datamodule(datamodule_config)

        with open(save_directory / DATAMODULE_CONFIG_NAME, "w") as outfile:
            yaml.dump(datamodule_config, outfile, sort_keys=False, default_flow_style=False)
        
        # Save the full experimental config
        if experiment_config_path is not None:
            shutil.copyfile(experiment_config_path, save_directory / FULL_CONFIG_NAME)

        card = self.create_hugging_face_model_card(card_template_path, wandb_repo, wandb_id)

        (save_directory / MODEL_CARD_NAME).write_text(str(card))

        if push_to_hub:
            api = HfApi()

            api.upload_folder(
                repo_id=hf_repo_id,
                folder_path=save_directory,
                repo_type="model",
                commit_message=f"Upload model - {wandb_id}",
            )

            # Print the most recent commit hash
            c = api.list_repo_commits(repo_id=hf_repo_id, repo_type="model")[0]

            message = (
                f"The latest commit is now: \n"
                f"    date: {c.created_at} \n"
                f"    commit hash: {c.commit_id}\n"
                f"    by: {c.authors}\n"
                f"    title: {c.title}\n"
            )

            print(message)

    @staticmethod
    def create_hugging_face_model_card(
        card_template_path: str,
        wandb_repo: str,
        wandb_id: str,
    ) -> ModelCard:
        """
        Creates Hugging Face model card

        Args:
            card_template_path: Path to the HuggingFace model card template
            wandb_repo: Identifier of the repo on wandb.
            wandb_id: Identifier of the model on wandb.

        Returns:
            card: ModelCard - Hugging Face model card object
        """

        # Creating and saving model card.
        card_data = ModelCardData(language="en", license="mit", library_name="pytorch")

        link = f"https://wandb.ai/{wandb_repo}/runs/{wandb_id}"
        wandb_link = f" - [{link}]({link})\n"

        # Find package versions for OCF packages
        packages_to_display = ["pvnet_summation", "ocf-data-sampler"]
        packages_and_versions = {package: version(package) for package in packages_to_display}


        package_versions_markdown = ""
        for package, v in packages_and_versions.items():
            package_versions_markdown += f" - {package}=={v}\n"

        return ModelCard.from_template(
            card_data,
            template_path=card_template_path,
            wandb_link=wandb_link,
            package_versions=package_versions_markdown,
        )


class BaseModel(torch.nn.Module, HuggingfaceMixin):
    """Abstract base class for PVNet-summation submodels"""

    def __init__(
        self,
        output_quantiles: list[float] | None,
        num_input_locations: int,
        input_quantiles: list[float] | None,
        history_minutes: int,
        forecast_minutes: int,
        interval_minutes: int,
    ):
        """Abtstract base class for PVNet-summation submodels.

        """
        super().__init__()

        if (output_quantiles is not None):
            if output_quantiles != sorted(output_quantiles):
                raise ValueError("output_quantiles should be in ascending order")
            if 0.5 not in output_quantiles:
                raise ValueError("Quantiles must include 0.5")

        self.output_quantiles = output_quantiles

        self.num_input_locations = num_input_locations
        self.input_quantiles = input_quantiles

        self.history_minutes = history_minutes
        self.forecast_minutes = forecast_minutes
        self.interval_minutes = interval_minutes

        # Number of timestemps for 30 minutely data
        self.history_len = history_minutes // interval_minutes
        self.forecast_len = (forecast_minutes) // interval_minutes

        # Store whether the model should use quantile regression or simply predict the mean
        self.use_quantile_regression = self.output_quantiles is not None

        # Also store the final output shape
        if self.use_quantile_regression:
            self.output_shape = (self.forecast_len, len(input_quantiles))
        else:
            self.output_shape = (self.forecast_len,)

        # Store the number of output features and that the model should predict for
        self.num_output_features = prod(self.output_shape)

        # Store the expected input shape
        if input_quantiles is None:
            self.input_shape = (self.num_input_locations, self.forecast_len)
        else:
            self.input_shape = (self.num_input_locations, self.forecast_len, len(input_quantiles))
            

    def _quantiles_to_prediction(self, y_quantiles: torch.Tensor) -> torch.Tensor:
        """Convert network prediction into a point prediction.

        Args:
            y_quantiles: Quantile prediction of network

        Returns:
            torch.Tensor: Point prediction
        """
        # y_quantiles Shape: [batch_size, seq_length, num_quantiles]
        idx = self.output_quantiles.index(0.5)
        return y_quantiles[..., idx]

    def sum_of_locations(self, x: SumTensorBatch) -> torch.Tensor:
        """Compute the sum of the location-level predictions"""
        if self.input_quantiles is None:
            y_hat = x["pvnet_outputs"]
        else:
            idx = self.input_quantiles.index(0.5)
            y_hat = x["pvnet_outputs"][..., idx]

        return (y_hat * x["relative_capacity"].unsqueeze(-1)).sum(dim=1)


"""Load a model from its checkpoint directory"""

import glob
import os

import hydra
import torch
import yaml

from pvnet_summation.utils import (
    DATAMODULE_CONFIG_NAME,
    FULL_CONFIG_NAME,
    MODEL_CONFIG_NAME,
)


def get_model_from_checkpoints(
    checkpoint_dir_path: str,
    val_best: bool = True,
) -> tuple[torch.nn.Module, dict, str | None, str | None]:
    """Load a model from its checkpoint directory

    Args:
        checkpoint_dir_path: str path to the directory with the model files
        val_best (optional): if True, load the best epoch model; otherwise, load the last

    Returns:
        tuple:
            model: nn.Module of pretrained model.
            model_config: dict of model config used to train the model.
            datamodule_config: path to datamodule used to create samples e.g train/test split info.
            experiment_configs: path to the full experimental config.

    """

    # Load lightning training module
    with open(f"{checkpoint_dir_path}/{MODEL_CONFIG_NAME}") as cfg:
        model_config = yaml.load(cfg, Loader=yaml.FullLoader)

    lightning_module = hydra.utils.instantiate(model_config)

    if val_best:
        # Only one epoch (best) saved per model
        files = glob.glob(f"{checkpoint_dir_path}/epoch*.ckpt")
        if len(files) != 1:
            raise ValueError(
                f"Found {len(files)} checkpoints @ {checkpoint_dir_path}/epoch*.ckpt. Expected one."
            )
        
        checkpoint = torch.load(files[0], map_location="cpu", weights_only=True)
    else:
        checkpoint = torch.load(
            f"{checkpoint_dir_path}/last.ckpt", 
            map_location="cpu", 
            weights_only=True,
        )

    lightning_module.load_state_dict(state_dict=checkpoint["state_dict"])

    # Extract the model from the lightning module
    model = lightning_module.model
    model_config = model_config["model"]

    # Check for datamodule config
    # This only exists if the model was trained with presaved samples
    datamodule_config = f"{checkpoint_dir_path}/{DATAMODULE_CONFIG_NAME}"
    datamodule_config = datamodule_config if os.path.isfile(datamodule_config) else None

    # Check for experiment config
    # For backwards compatibility - this might not always exist
    experiment_config = f"{checkpoint_dir_path}/{FULL_CONFIG_NAME}"
    experiment_config = experiment_config if os.path.isfile(experiment_config) else None

    return model, model_config, datamodule_config, experiment_config



"""Pytorch lightning datamodules for loading pre-saved samples and predictions."""

import os
from glob import glob
from typing import TypeAlias

import numpy as np
import pandas as pd
import torch
from lightning.pytorch import LightningDataModule
from ocf_data_sampler.load.generation import open_generation
from ocf_data_sampler.numpy_sample.common_types import NumpyBatch
from ocf_data_sampler.numpy_sample.sun_position import calculate_azimuth_and_elevation
from ocf_data_sampler.torch_datasets.pvnet_dataset import PVNetConcurrentDataset
from ocf_data_sampler.utils import minutes
from torch.utils.data import DataLoader, Dataset, Subset, default_collate
from typing_extensions import override

SumNumpySample: TypeAlias = dict[str, np.ndarray | NumpyBatch]
SumTensorBatch: TypeAlias = dict[str, torch.Tensor]


def construct_sample(
    pvnet_inputs: NumpyBatch,
    valid_times: pd.DatetimeIndex,
    relative_capacities: np.ndarray,
    target: np.ndarray | None,
    longitude: float,
    latitude: float,
    last_outturn: float | None = None,
) -> SumNumpySample:
    """Construct an input sample for the summation model

    Args:
        pvnet_inputs: The PVNet batch for all locations
        valid_times: An array of valid-times for the forecast
        relative_capacities: Array of capacities of all locations normalised by the total capacity
        target: The target national outturn. This is only needed during training.
        longitude: The longitude of the national centroid
        latitude: The latitude of the national centroid
        last_outturn: The previous national outturn. This is only needed during training.

    """

    azimuth, elevation = calculate_azimuth_and_elevation(valid_times, longitude, latitude)

    sample = {
        # NumpyBatch object with batch size = num_locations
        "pvnet_inputs": pvnet_inputs,
        # Shape: [time]
        "valid_times": valid_times.values.astype(int),
        # Shape: [num_locations]
        "relative_capacity": relative_capacities,
        # Shape: [time]
        "azimuth": azimuth.astype(np.float32) / 360,
        # Shape: [time]
        "elevation": elevation.astype(np.float32) / 180 + 0.5,
    }

    if target is not None:
        # Shape: [time]
        sample["target"] = target
    if last_outturn is not None:
        # Shape: scalar
        sample["last_outturn"] = last_outturn
    return sample


class StreamedDataset(PVNetConcurrentDataset):
    """A torch dataset for creating concurrent PVNet inputs and national targets."""

    def __init__(
        self,
        config_filename: str,
        start_time: str | None = None,
        end_time: str | None = None,
    ) -> None:
        """A torch dataset for creating concurrent PVNet inputs and national targets.

        Args:
            config_filename: Path to the configuration file
            start_time: Limit the init-times to be after this
            end_time: Limit the init-times to be before this
        """
        super().__init__(config_filename, start_time, end_time)

        self.national_data = (
            open_generation(
                zarr_path=self.config.input_data.generation.zarr_path,
            )
            .sel(location_id=0)
            .compute()
        )

        self.longitude = self.national_data.longitude.item()
        self.latitude = self.national_data.latitude.item()

    def _get_sample(self, t0: pd.Timestamp) -> SumNumpySample:
        """Generate a concurrent PVNet sample for given init-time.

        Args:
            t0: init-time for sample
        """

        # Get the PVNet input batch
        pvnet_inputs: NumpyBatch = super()._get_sample(t0)

        # Construct an array of valid times for each forecast horizon
        valid_times = pd.date_range(
            t0 + minutes(self.config.input_data.generation.time_resolution_minutes),
            t0 + minutes(self.config.input_data.generation.interval_end_minutes),
            freq=minutes(self.config.input_data.generation.time_resolution_minutes),
        )

        # Get the region and national capacities
        location_capacities = pvnet_inputs["capacity_mwp"]
        total_capacity = self.national_data.sel(time_utc=t0).capacity_mwp.item()

        # Calculate requited inputs for the sample
        relative_capacities = location_capacities / total_capacity
        target = self.national_data.sel(time_utc=valid_times).values / total_capacity
        last_outturn = self.national_data.sel(time_utc=t0).values / total_capacity

        return construct_sample(
            pvnet_inputs=pvnet_inputs,
            valid_times=valid_times,
            relative_capacities=relative_capacities,
            target=target,
            longitude=self.longitude,
            latitude=self.latitude,
            last_outturn=last_outturn,
        )

    @override
    def __getitem__(self, idx: int) -> SumNumpySample:
        return super().__getitem__(idx)

    @override
    def get_sample(self, t0: pd.Timestamp) -> SumNumpySample:
        return super().get_sample(t0)


class StreamedDataModule(LightningDataModule):
    """Datamodule for training pvnet_summation."""

    def __init__(
        self,
        configuration: str,
        train_period: list[str | None] = [None, None],
        val_period: list[str | None] = [None, None],
        num_workers: int = 0,
        prefetch_factor: int | None = None,
        persistent_workers: bool = False,
        seed: int | None = None,
        dataset_pickle_dir: str | None = None,
    ):
        """Datamodule for creating concurrent PVNet inputs and national targets.

        Args:
            configuration: Path to ocf-data-sampler configuration file.
            train_period: Date range filter for train dataloader.
            val_period: Date range filter for val dataloader.
            num_workers: Number of workers to use in multiprocess batch loading.
            prefetch_factor: Number of data will be prefetched at the end of each worker process.
            persistent_workers: If True, the data loader will not shut down the worker processes
                after a dataset has been consumed once. This allows to maintain the workers Dataset
                instances alive.
            seed: Random seed used in shuffling datasets.
            dataset_pickle_dir: Directory in which the val and train set will be presaved as
                pickle objects. Setting this speeds up instantiation of multiple workers a lot.
        """
        super().__init__()
        self.configuration = configuration
        self.train_period = train_period
        self.val_period = val_period
        self.seed = seed
        self.dataset_pickle_dir = dataset_pickle_dir

        self._dataloader_kwargs = dict(
            batch_size=None,
            batch_sampler=None,
            num_workers=num_workers,
            collate_fn=None,
            pin_memory=False,
            drop_last=False,
            timeout=0,
            worker_init_fn=None,
            prefetch_factor=prefetch_factor,
            persistent_workers=persistent_workers,
            multiprocessing_context="spawn" if num_workers > 0 else None,
        )

    def setup(self, stage: str | None = None):
        """Called once to prepare the datasets."""

        # This logic runs only once at the start of training, therefore the val dataset is only
        # shuffled once
        if self.dataset_pickle_dir is not None:
            os.makedirs(self.dataset_pickle_dir, exist_ok=True)

            train_dataset_path = f"{self.dataset_pickle_dir}/train_dataset.pkl"
            val_dataset_path = f"{self.dataset_pickle_dir}/val_dataset.pkl"

            # For safety, these pickled datasets cannot be overwritten.
            # See: https://github.com/openclimatefix/pvnet/pull/445
            for path in [train_dataset_path, val_dataset_path]:
                if os.path.exists(path):
                    raise FileExistsError(
                        f"The pickled dataset path '{path}' already exists. Make sure that "
                        "this can be safely deleted (i.e. not currently being used by any "
                        "training run) and delete it manually. Else change the "
                        "`dataset_pickle_dir` to a different directory."
                    )

        # Prepare the train dataset
        self.train_dataset = StreamedDataset(self.configuration, *self.train_period)

        # Prepare and pre-shuffle the val dataset and set seed for reproducibility
        val_dataset = StreamedDataset(self.configuration, *self.val_period)
        shuffled_indices = np.random.default_rng(seed=self.seed).permutation(len(val_dataset))
        self.val_dataset = Subset(val_dataset, shuffled_indices)

        if self.dataset_pickle_dir is not None:
            self.train_dataset.presave_pickle(train_dataset_path)
            self.train_dataset.presave_pickle(val_dataset_path)

    def teardown(self, stage: str | None = None) -> None:
        """Clean up the pickled datasets"""
        if self.dataset_pickle_dir is not None:
            for filename in ["val_dataset.pkl", "train_dataset.pkl"]:
                filepath = f"{self.dataset_pickle_dir}/{filename}"
                if os.path.exists(filepath):
                    os.remove(filepath)

    def train_dataloader(self, shuffle: bool = False) -> DataLoader:
        """Construct train dataloader"""
        return DataLoader(self.train_dataset, shuffle=shuffle, **self._dataloader_kwargs)

    def val_dataloader(self, shuffle: bool = False) -> DataLoader:
        """Construct val dataloader"""
        return DataLoader(self.val_dataset, shuffle=shuffle, **self._dataloader_kwargs)


class PresavedDataset(Dataset):
    """Dataset for loading pre-saved PVNet predictions from disk"""

    def __init__(self, sample_dir: str):
        """Dataset for loading pre-saved PVNet predictions from disk.

        Args:
            sample_dir: The directory containing the saved samples
        """
        self.sample_filepaths = sorted(glob(f"{sample_dir}/*.pt"))

    def __len__(self) -> int:
        return len(self.sample_filepaths)

    def __getitem__(self, idx: int) -> dict:
        return torch.load(self.sample_filepaths[idx], weights_only=True)


class PresavedDataModule(LightningDataModule):
    """Datamodule for loading pre-saved PVNet predictions."""

    def __init__(
        self,
        sample_dir: str,
        batch_size: int = 16,
        num_workers: int = 0,
        prefetch_factor: int | None = None,
        persistent_workers: bool = False,
    ):
        """Datamodule for loading pre-saved PVNet predictions.

        Args:
            sample_dir: Path to the directory of pre-saved samples.
            batch_size: Batch size.
            num_workers: Number of workers to use in multiprocess batch loading.
            prefetch_factor: Number of data will be prefetched at the end of each worker process.
            persistent_workers: If True, the data loader will not shut down the worker processes
                after a dataset has been consumed once. This allows to maintain the workers Dataset
                instances alive.
        """
        super().__init__()
        self.sample_dir = sample_dir

        self._dataloader_kwargs = dict(
            batch_size=batch_size,
            sampler=None,
            batch_sampler=None,
            num_workers=num_workers,
            collate_fn=None if batch_size is None else default_collate,
            pin_memory=False,
            drop_last=False,
            timeout=0,
            worker_init_fn=None,
            prefetch_factor=prefetch_factor,
            persistent_workers=persistent_workers,
            multiprocessing_context="spawn" if num_workers > 0 else None,
        )

    def train_dataloader(self, shuffle: bool = True) -> DataLoader:
        """Construct train dataloader"""
        dataset = PresavedDataset(f"{self.sample_dir}/train")
        return DataLoader(dataset, shuffle=shuffle, **self._dataloader_kwargs)

    def val_dataloader(self, shuffle: bool = False) -> DataLoader:
        """Construct val dataloader"""
        dataset = PresavedDataset(f"{self.sample_dir}/val")
        return DataLoader(dataset, shuffle=shuffle, **self._dataloader_kwargs)
